// show-signal-stack.c, 64 Bit
// 
// Systemprogrammierung, SS 2013
// Hans-Georg Esser

#include <stdio.h>    // printf, snprintf
#include <unistd.h>   // sleep, alarm
#include <stdlib.h>   // exit, system
#include <signal.h>   // signal (), SIGALRM

// Makro-Definitionen
#define get_esp(v) __asm__ ("mov %%rsp, %0" : "=r"(v))
#define get_ebp(v) __asm__ ("mov %%rbp, %0" : "=r"(v))

#define read_memory(addr) (*(unsigned long*)(addr))
#define write_memory(addr, value) (*(unsigned long *)(addr) = ((unsigned long)value))

#define print_stack(name,adr,count) {  \
  int i;                               \
  for (i=0; i<count; i++) {            \
    printf (#name ": [%016lx] = [esp+%4d] = %016lx\n", \
      adr+i*8, i*8, read_memory(adr+i*8));         \
    fflush (stdout);                   \
  }                                    \
}

#define print_regs(name) \
  printf (#name ": esp = 0x%016lx, ebp = 0x%016lx, eip = 0x%016lx\n", \
    esp, ebp, eip);

void get_eip (unsigned long *v) {
  unsigned long eip;
  __asm__ __volatile__ ("movq 8(%%rbp), %0" : "=r" (eip));
  *v = eip;
}

#define get_regs get_esp (esp); get_ebp (ebp); get_eip (&eip)

void jump_test ();

void alarm_handler (int sig) {
  unsigned long esp, ebp, eip;
  printf ("alrm: sig = %d\n", sig);

  // alarm: Register
  get_regs;
  print_regs (alrm);
  print_stack (alrm, esp, 50);
  
  // Hier aendern wir einen Wert auf dem Stack:
  // esp zeigt auf das Ende vom Stack,
  // &jump_test ist die Adresse der Funktion jump_test().
  write_memory (esp+248, &jump_test);
  return;
}

void jump_test () {
  printf ("ENTERED jump_test()\n");
  for (;;);
};

int main () {
  unsigned long esp, ebp, eip;
  char s[100];  // for command string
  int pid = getpid ();
  printf ("pid = %d\n", pid);
  printf ("&main = %p\n", main);

  signal (SIGALRM, alarm_handler);  // alarm handler
  alarm (1);                        // alarm nach 1 sek.

  get_eip (&eip);
  printf ("main: eip = %lx\n", eip);

  for (;;) ;
}
