#define _GNU_SOURCE 1
#include <signal.h>
#include <ucontext.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

// Makro kopiert Inhalt vom RSP-Register in Long-Variable v
#define get_rsp(v) __asm__ ("mov %%rsp, %0" : "=r"(v))

// Makro liest 64-Bit-Wert (long) ab Adresse addr und gibt ihn
// als Ergebnis zurueck
#define PEEK(addr) (*(unsigned long*)(addr))

// Moderne Timer mit Sekunden und Mikrosekunden
struct itimerval timer;

void jump_test () {
  printf ("ENTERED jump_test()\n");
  for (;;);
};

static void alarm_handler(int sig, siginfo_t *si, void *ptr) {
  printf ("Im Signal Handler\n");
  ucontext_t *uc = (ucontext_t *)ptr;
  unsigned long adr = uc->uc_mcontext.gregs[REG_RIP];

  // aktueller Stack:
  unsigned long rsp;
  get_rsp (rsp);
  printf ("  rsp             = 0x%lx\n", rsp);
  // Wo liegt gregs?
  printf ("  &gregs          = %p\n", &uc->uc_mcontext.gregs);
  printf ("  &gregs[REG_RIP] = %p\n", &uc->uc_mcontext.gregs[REG_RIP]);
  // Und das RIP-Feld enthaelt ...
  printf ("  gregs[REG_RIP]  = 0x%lx\n", adr);
    
  // Neue Ruecksprungadresse eintragen
  adr = (unsigned long)&jump_test;
  printf ("  Neue Adresse: 0x%lx (= &jump_test)\n", adr);  
  uc->uc_mcontext.gregs[REG_RIP] = adr;
  
  printf ("Verlasse Signal Handler\n");
}


#define STACKSIZE 0x4000

int main() {
  // Handler eintragen, moderne Methode mit sigaction
  struct sigaction s;
  s.sa_flags = 0;
  s.sa_sigaction = alarm_handler;
  sigemptyset(&s.sa_mask);
  sigaction(SIGVTALRM, &s, NULL);

  // Timer setzen
  timer.it_value.tv_sec     = 0;
  timer.it_value.tv_usec    = 10000;   // nach 10 Millisek.
  timer.it_interval.tv_sec  = 0;
  timer.it_interval.tv_usec = 0;       // keine Wiederholung
  setitimer (ITIMER_VIRTUAL, &timer, NULL);

  for (;;) ;         // Endlosschleife im Hauptprogramm
}

