// 19c-threads-02.c, 64 Bit
//
// Systemprogrammierung, SS 2013
// Hans-Georg Esser

#define _GNU_SOURCE 1
#include <signal.h>
#include <ucontext.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

/*   ucontext_t = struct ucontext:
                 /usr/include/asm-generic/ucontext.h
     - enthaelt: mcontext_t uc_mcontext

     mcontext_t: /usr/include/asm-generic/ucontext.h
                 /usr/include/x86_64-linux-gnu/sys/ucontext.h
     - enthaelt: gregset_t gregs;
                 typedef greg_t gregset_t[NGREG];

     Register in /usr/include/x86_64-linux-gnu/sys/ucontext.h   */

#define QUANTUM   100000     // Quantum in mikrosec. fuer Scheduler
#define STACKSIZE 0x4000     // Groesse des Thread-Stacks
#define MAXTHREADS 20        // bis zu 20 Threads

// Globale Variablen fuer Thread-Verwaltung
int            tid = -1;            // noch kein Thread
int            thread_count = 0;    // Anzahl der Threads

// Thread-spezifische Daten
unsigned long  rbps[MAXTHREADS];      // RBPs (nur fuer Start)
unsigned long  rsps[MAXTHREADS];      // RSPs (nur fuer Start)
unsigned long  start[MAXTHREADS];     // Einsprungadressen
mcontext_t     context[MAXTHREADS];   // Kontexte
short          firstrun[MAXTHREADS];  // schon gelaufen?
void*          stacks[MAXTHREADS];    // Thread-Stacks

// Sonstige Variablen
long           c1, c2, c3;            // globale Counter

// Thread initialisieren
void create_thread (int tid, void *(*function)(void *)) {
  // tid: Thread-ID
  // function: Adresse der Thread-Funktion (also void* f(...) definiert)
  if (tid != thread_count || tid >= MAXTHREADS) return;  // Fehler
  
  firstrun[tid] = 1;                      // Thread wurde noch nie ausgefuehrt

  // TO DO:
  // - Speicher fuer (einen) neuen Stack reservieren und in stacks[] merken
  // - Einsprungadresse (  (unsigned long)function  ) in start[] merken
  // - Eintraege in rsps[], rbps[] wie vorher, Stack beginnt am Ende!

  printf ("new thread: %lx\n", start[tid]);
  
  thread_count = tid+1;  // Das funktioniert nur, wenn wir die Threads in
                         // richtiger Reihenfolge (0, 1, 2, ...) erzeugen
}

// Thread-Funktionen
long int rechnen (long x, long y) { return x*3+y; }

void *test01 () {
  long count;
  for (;;) {
    printf ("HALLO WELT [%ld], tid = %d -- TESTFUNKTION, CALC = %ld\n", 
      count++, tid, rechnen (count,tid) );
    c1 = count;
    usleep (10000);
  };
}

void *test02 () {
  long count;
  for (;;) {
    printf ("hallo welt [%ld], tid = %d -- zweiter test\n", count++, tid);
    c2 = count;
    usleep (10000);
  };
}

void *vor_test02 () { test02 (); }
void *vor_vor_test02 () { vor_test02 (); }

void *test03 () {
  long int count;
  for (;;) {
    printf ("yo hey wha [%ld], tid = %d -- dritter test\n", count++, tid);
    c3 = count;
    usleep (10000);
  };
}

// Signal-Handler fuer das Timer-Signal
static void timer_handler(int sig, siginfo_t *si, void *ptr) {
  ucontext_t *uc = (ucontext_t *)ptr;

  // Aktuelle Werte (alter Thread) auslesen
  // (wird nur fuer die Debug-Ausgabe hier verwendet)
  unsigned long rip = uc->uc_mcontext.gregs[REG_RIP];
  unsigned long rsp = uc->uc_mcontext.gregs[REG_RSP];
  unsigned long rbp = uc->uc_mcontext.gregs[REG_RBP];
  printf ("\nrip: %lx, rsp: %lx, rbp: %lx\n", rip, rsp, rbp);
  
  // aktuellen Kontext sichern
  if (tid != -1) {  // falls schon ein Thread lief
  
    // TO DO: Aktuellen Kontext uc->uc_mcontext sichern
    
  }
  
  // Neuen Thread auswaehlen (es gibt nur TIDs 0, 1, ..., thread_count-1)
  if (tid != -1)
    tid = (tid+1) % thread_count;   // naechste Thread-ID bestimmen
  else
    tid = 0;       // mit Thread 0 starten
  printf ("Setze tid auf %d. ", tid);
  
  if (firstrun[tid] == 1) {
    printf ("Starte thread %d. ", tid);
  }
  
  printf ("Neue Adresse: %lx, rsp: %lx, rbp: %lx\n", rip, rsps[tid], rbps[tid]);
  
  if (firstrun[tid] == 1) {
    firstrun[tid] = 0;
    // Manuelles Register-Setzen nur beim Thread-Start, spaeter ueber Kontext

    // TO DO: Werte aus start[], rsps[], rbps[] in die richtigen
    //        gregs-Felder eintragen

  } else {
  
    // TO DO: Kontext zurueck sichern

  }
}

// Signal-Handler fuer Strg-C
static void ctrl_c_handler (int sig) {
  printf ("\nCtrl-c\n" "c1 = %ld, c2 = %ld, c3 = %ld\n", c1, c2, c3);
  exit (0);
}

int main() {
  // Threads vorbereiten, tid muss in aufsteigender Reihenfolge sein
  create_thread (0, test01);
  create_thread (1, vor_vor_test02);
  create_thread (2, vor_vor_test02);
  create_thread (3, test03);
  create_thread (4, test03);

  // Alarm-Signal-Handler eintragen
  struct sigaction s;
  s.sa_flags = SA_RESTART;
  s.sa_sigaction = timer_handler;
  sigemptyset(&s.sa_mask);
  sigaction(SIGALRM, &s, NULL);

  // Signal-Handler (Strg-C) eintragen
  signal (SIGINT, ctrl_c_handler);

  // Timer setzen
  struct itimerval timer;
  timer.it_value.tv_sec    = 0; timer.it_value.tv_usec    = QUANTUM;
  timer.it_interval.tv_sec = 0; timer.it_interval.tv_usec = QUANTUM;
  setitimer (ITIMER_REAL, &timer, NULL);

  // Endlosschleife
  for (;;) ;
}
